use std::{
    collections::HashMap,
    sync::{Arc, Mutex},
};

use anyhow::Result;
use utilities::color::Color;
use vulkan_rs::prelude::*;

#[derive(Debug, Clone, Copy)]
pub enum ElementBorderThickness {
    Pixel(u32),
    Ratio(f32),
}

impl From<u32> for ElementBorderThickness {
    fn from(value: u32) -> Self {
        Self::Pixel(value)
    }
}

impl From<f32> for ElementBorderThickness {
    fn from(value: f32) -> Self {
        Self::Ratio(value)
    }
}

#[derive(Debug, Clone, Copy)]
pub struct ElementDescriptor {
    pub background_color: Color,
    pub border_color: Color,

    pub border_thickness: ElementBorderThickness,
}

impl ElementDescriptor {
    pub fn new(
        background_color: impl Into<Color>,
        border_color: impl Into<Color>,
        border_thickness: impl Into<ElementBorderThickness>,
    ) -> Self {
        Self {
            background_color: background_color.into(),
            border_color: border_color.into(),
            border_thickness: border_thickness.into(),
        }
    }
}

#[derive(Debug, Hash, Clone, Copy, PartialEq, Eq)]
struct ElementDefinition {
    width: u32,
    height: u32,

    background_color: Color,
    border_color: Color,
    border_thickness: u32,
}

pub struct ElementCreator {
    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,

    elements: HashMap<ElementDefinition, Arc<Image>>,
}

impl ElementCreator {
    pub fn new(device: Arc<Device>, queue: Arc<Mutex<Queue>>) -> Self {
        Self {
            device,
            queue,
            elements: HashMap::new(),
        }
    }

    pub fn get(
        &mut self,
        width: u32,
        height: u32,
        descriptor: ElementDescriptor,
    ) -> Result<Arc<Image>> {
        let definition = ElementDefinition {
            width,
            height,

            background_color: descriptor.background_color,
            border_color: descriptor.border_color,

            border_thickness: match descriptor.border_thickness {
                ElementBorderThickness::Pixel(pixel) => pixel,
                ElementBorderThickness::Ratio(ratio) => {
                    (width.min(height) as f32 * ratio).round() as u32
                }
            },
        };

        match self.elements.get(&definition) {
            Some(element) => Ok(element.clone()),
            None => {
                let element = self.create_element(definition)?;
                self.elements.insert(definition, element.clone());

                Ok(element)
            }
        }
    }

    fn create_element(&self, definition: ElementDefinition) -> Result<Arc<Image>> {
        todo!()
    }
}

struct Generator {
    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,

    max_supported_sample_count: VkSampleCountFlags,
    vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
    fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
}

impl Generator {
    fn new(device: Arc<Device>, queue: Arc<Mutex<Queue>>) -> Result<Self> {
        Ok(Self {
            max_supported_sample_count: device.max_supported_sample_count(VK_SAMPLE_COUNT_16_BIT),

            vertex_shader: ShaderModule::from_slice(
                device.clone(),
                include_bytes!("generator.vert.spv"),
            )?,
            fragment_shader: ShaderModule::from_slice(
                device.clone(),
                include_bytes!("generator.frag.spv"),
            )?,

            device,
            queue,
        })
    }

    fn generate(&self, definition: ElementDefinition) -> Result<Arc<Image>> {
        let image = self.create_image(definition.width, definition.height)?;

        todo!()
    }

    fn create_image(&self, width: u32, height: u32) -> Result<Arc<Image>> {
        let image = Image::empty(
            width,
            height,
            VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT,
            VK_SAMPLE_COUNT_1_BIT,
        )
        .format(VK_FORMAT_R8G8B8A8_UNORM)
        .attach_sampler(Sampler::nearest_sampler().build(&self.device)?)
        .build(&self.device, &self.queue)?;

        image.convert_layout(VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL)?;

        Ok(image)
    }

    fn render_target(&self, image: &Arc<Image>) -> Result<RenderTarget> {
        RenderTarget::builder()
            .add_sub_pass(
                SubPass::builder(image.width(), image.height())
                    .add_target_info(CustomTarget {
                        usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT.into(),
                        format: image.vk_format(),
                        clear_on_load: true,
                        store_on_save: true,
                        attach_sampler: false,
                        use_as_input: false,
                        clear_value: ClearValue::Color([0.0, 0.0, 0.0, 0.0]),
                    })
                    .set_sample_count(self.max_supported_sample_count)
                    .add_resolve_targets((vec![image.clone()], true))
                    .use_queue(image.queue().clone())
                    .build(image.device())?,
            )
            .build(image.device())
    }
}
